/*! \mainpage Speeded Up Speeded Up Robust Features
 *
 * A GPU implementation of the SURF algorithm
 *
 * \image html surf-example-mars.png "Image credit: NASA/JPL/University of Arizona. Image from the edge of the south polar residual cap on Mars." 
 *
 * \section toc Table of Contents
 * <ul>
 *  <li> \ref sec-intro </li>
 *  <li> \ref sec-details </li>
 *  <ul>
 *    <li> \ref ssec-lib </li>
 *    <li> \ref ssec-cline </li>
 *    <li> \ref ssec-matlab </li>
 *  </ul>
 *  <li> \ref sec-known-deviations </li>
 *  <li> \ref sec-limitations </li>
 *  <li> \ref sec-build </li>
 *   <ul>
 *     <li>\ref page-build-windows </li>
 *     <li>\ref page-build-osx</li>
 *     <li>\ref page-build-linux</li>
 *   </ul>
 * <li>\ref sec-download</li>
 * <li>\ref sec-projects</li>
 * <li>\ref sec-authors</li>
 * <li>\ref sec-contact</li>
 * <li>\ref sec-license</li>
 * <li>\ref sec-acknowledgements</li>
 * <li>\ref sec-references</li>
 * </ul>
 * \section sec-intro Introduction
 *
 *
 * This is an implementation of the SURF algorithm [1] using the NVIDIA CUDA API [2] and released under a BSD license.
 *
 * In implementing this algorithm, the focus was on accuracy, not speed. That is, we attempted to match the results of SURF where possible, and where not, we attempted our best guess. Additionally, a parallel Matlab implementation is available, which is useful for understanding the algorithm and the implementation.
 * Further details on the implementation (some out of date) are available as a report on the original code <A HREF=http://asrl.utias.utoronto.ca/~ptf/docs/gpusurf_report09.pdf>[pdf]</A>.
 *
 * \section sec-details Usage
 *
 * \subsection ssec-lib As a library
 * 
 * Compiling the source code provides a dynamic link library (.dll in Windows, .dylib in OSX, .so in Linux). High-level use of the library is provided through the \link asrl::GpuSurfDetector GpuSurfDetector \endlink object. An example demonstrating all functionality of detector object is provided in the file gpusurf_engine.cpp. 
 *
 * \subsection ssec-cline As a command line utility
 * The gpusurf_engine example program can be run through a command line interface. Help on syntax is provided if the program is run without any input options. An illustration of how the input filter parameters are used can be found in the documentation for the \link asrl::GpuSurfOctave GpuSurfOctave\endlink class. For a series of images, all of the image file names should be provided in a single command. This is faster than processing each image sequentially, as re-initialization is not required between images. 
 *
 * The output for each image is a .key text file, which consists of all of the keypoints and their associated descriptors in each row. For visualization purposes, these .key files can be loaded into Matlab using the load_gpu_keypoints() function.
 *
 * \subsection ssec-matlab From within Matlab
 * The parallel implementation coded in Matlab can be run by using the surf_find_keypoints() function. The output keypoints can be sorted by strength using surf_best_n_keypoints(), and plotted using surf_plot_keypoints().
 *
 * \section sec-known-deviations Deviations from SURF
 * Due to the closed-source nature of the original SURF distribution [1], we were not able to reproduce exactly matching results with our algorithm. However, it is to the best of our knowledge that no other available open-source implementations have been able to accomplish this task either. The results of the Mikolajczyk [3] repeatability test for the Graffiti dataset for our implementation and other available detectors can be seen below.
 *
 * \image html miko-graffiti.png
 *
 * The deviations we have identified are:
 * <ul>
 *	<li> The GPU hardware produces slightly different computational results due to the single precision floats.
 *	<li> Due to the custom-dimensioned box filters, the reported keypoint scale is somewhat larger than what SURF produces for a similarly sized blob. For the default parameters, it is approximately 1.3 times larger. This larger value causes the orientation and descriptors are computed at a larger scale.
 *	<li> The Dxy filter gap grows with scale, unlike the illustration in [1].
 *	<li> The computed descriptors use a Gaussian weighting of 3.3s as described in [1], but descriptors obtained from SURF seem to indicate another value is used.
 * </ul>
 *
 * \section sec-limitations Known limitations
 * <ul>
 *	<li> The default GPU thread counts should be optimized for specific hardware.
 *	<li> Inability to select the device for multi-GPU systems.
 *	<li> Integral image computation can likely be sped up by using the 2D parallel prefix sum by Terriberry et al. [4].
 * </ul>
 *
 * \section sec-build Building the library on different platforms
 * The bulk of our development has been on MacBook Pro computers that dual boot into Windows XP. As such, the build instructions are best for those platforms. However, no platform-specific libraries were used and so compiling on other platforms should be straigtforward, provided they are supported by CUDA. The main difficulty with compiling the library arises from dependencies: OpenCV, Boost, and CUDPP. We have included the source of CUDPP in our download and provided links and instructions for obtaining the other libraries for each platform. 
 *
 * \ref page-build-windows 
 *
 * \ref page-build-osx
 *
 * \subsection page-build-linux Build instructions for Linux
 * No one of us has a Linux box with a GPU. The basic build should be the same as OSX but packages (such as Boost, CMake, and OpenCV should come from your distribution's package manager (aptitude, yast, etc).
 *
 * \section sec-download Download
 * 
 * Release 0.1.03738 <A HREF="http://dstats.net/download/http://asrl.utias.utoronto.ca/code/gpusurf/gpusurf-0.1.03738.tgz">[download]</A> \ref page-release-notes "[release notes]"
 *
 * <ul>
 * <li>Number of times downloaded: 
 * \htmlonly
 	<script type='text/javascript' src='http://dstats.net/dstatsjs.php?file=http://asrl.utias.utoronto.ca/code/gpusurf/gpusurf-0.1.03738.tgz'>
	</script>
	<script type='text/javascript'>
 		document.write(dsCounter);
	</script>
	<a href='http://dstats.net/show/http://asrl.utias.utoronto.ca/code/gpusurf/gpusurf-0.1.03738.tgz'>[stats]</a>
 * \endhtmlonly
 * </li>
 *
 * \section sec-projects Projects using this code
 * <ul>
 *  <li> <A HREF="http://asrl.utias.utoronto.ca/~ptf/JFR_2009">Visual Teach and Repeat for Long-Range Rover Autonomy</A> </li>
 * </ul>
 *
 * Please send us links to your project!
 *
 * \section sec-authors About the Authors
 * Paul Furgale is a PhD candidate in the Autonomous Space Robotics Lab at the University of Toronto Institute for Aerospace Studies. His research works to expand the basic Visual Odometry pipeline to enable long-range autonomous rover navigation. <a href="http://asrl.utias.utoronto.ca/~ptf">[website]</a>
 *
 * Chi Hay Tong is a PhD student in the Autonomous Space Robotics Lab at the University of Toronto Institute for Aerospace Studies. His research interests involve robust 3D worksite mapping using laser rangefinders. <a href="http://asrl.utias.utoronto.ca/~cht">[website]</a>
 *
 * \section sec-contact Contact Information
 * Inquiries about the project can be directed at (paul.furgale_AT_NOSPAM_utoronto.ca).
 * 
 * \section sec-license License
 * This code is released under a BSD license.
 * 
 * \section sec-acknowledgements Acknowledgements
 * Many people have helped us in the development of this code and we would like to thank some of them here. Gaetan Kenway helped with the initial development of the code during an excellent course run by Dr. A. Moshovos at the University of Toronto. Thanks to Mikolajczyk et al. [3] and Bay et al. [1] for providing us with libraries that let us evaluate if we were on the right track. Finally, thanks to our supervisor Tim Barfoot for providing equipment and support, even though we were supposed to be working on our theses.
 *
 * \section sec-references References
 * [1] H. Bay, A. Ess, T. Tuytelaars, and L. Van Gool. Speeded-up robust features (SURF). \a Computer \a Vision \a and \a Image \a Understanding, 110(3):346-359, 2008.
 *
 * [2] NVIDIA Corporation, 2701 San Tomas Expressway, Santa Clara, CA. NVIDIA CUDA Programming Guide, 2.3.1 edition.
 *
 * [3] K. Mikolajczyk, T. Tuytelaars, C. Schmid, A. Zisserman, J. Matas, F. Schaffalitzky, T. Kadir, and L. V. Gool. A comparison of affine region detectors. \a International \a Journal \a of \a Computer \a Vision, 65(1-2):43-72, November 2005.
 *
 * [4] T. Terriberry, L. French, and J. Helmsen. GPU Accelerating Speeded-Up Robust Features. \a In \a Proceedings \a of \a the \a 4th \a International \a Symposium \a on \a 3D \a Data \a Processing, \a Visualization \a and \a Transmission, 355–362, June 2008.
 *
 */


