/*
Copyright (c) 2010, Paul Furgale and Chi Hay Tong
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are 
met:

* Redistributions of source code must retain the above copyright notice, 
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright 
  notice, this list of conditions and the following disclaimer in the 
  documentation and/or other materials provided with the distribution.
* The names of its contributors may not be used to endorse or promote 
  products derived from this software without specific prior written 
  permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#undef _GLIBCXX_ATOMIC_BUILTINS
#include <cudpp.h>
#include "assert_macros.hpp"


#define BLOCK_DIM 16

namespace asrl {
  // convert unsigned chars into floats (scaled by 255.0f)
  // - second function does nothing - overloaded, but __device__ functions are inline, so no overhead
  __device__ void convert_dev(float & out, unsigned char in){ out = (float) in / 255.0f; }
  __device__ void convert_dev(float & out, float in){ out = in; }


  // matrix transpose operation (on the GPU)
  template <typename T>
  __global__ void transpose_kernel(float *odata, size_t o_pitch, T *idata, size_t i_pitch, size_t width, size_t height)
  {
    __shared__ float block[BLOCK_DIM][BLOCK_DIM+1];

    // read the matrix tile into shared memory
    unsigned int xBlock = __mul24(blockDim.x, blockIdx.x);
    unsigned int yBlock = __mul24(blockDim.y, blockIdx.y);
    unsigned int xIndex = xBlock + threadIdx.x;
    unsigned int yIndex = yBlock + threadIdx.y;

    if ((xIndex < width) && (yIndex < height))
      {
	// load block into shared memory
	unsigned int index_in = __mul24(i_pitch, yIndex) + xIndex;	// where from in data
	convert_dev(block[threadIdx.y][threadIdx.x], idata[index_in]);	// convert to float (if not already)
      }

    __syncthreads();

    // write it back to global memory
    xIndex = yBlock + threadIdx.x;
    yIndex = xBlock + threadIdx.y;
    if ((xIndex < height) && (yIndex < width))
      {
	unsigned int index_out = __mul24(o_pitch, yIndex) + xIndex;
	odata[index_out] = block[threadIdx.x][threadIdx.y];
      }
  }




  void call_integral_kernel(size_t width, size_t height,
			    unsigned char * char_data,		// initial storage on the gpu of the unsigned char data
			    float *norm_data, 
			    float *trans_data,	// storage of floats on the gpu
			    cudaArray *output_int_img,
			    size_t norm_pitch, 
			    size_t trans_pitch, 
			    size_t char_pitch,
			    CUDPPHandle & rowPlan,
			    CUDPPHandle & colPlan,
			    cudaStream_t stream)
  {
    // transpose and convert the data into floats
    dim3 block(BLOCK_DIM, BLOCK_DIM, 1);
    dim3 grid( (int) ((block.x + width - 1) / block.x), (int) ((block.y + height - 1) / block.y), 1);

    //TIME_START("iimg: transpose 1");
    transpose_kernel<unsigned char> <<< grid, block, 0>>> (trans_data, trans_pitch/sizeof(float), char_data, char_pitch/sizeof(unsigned char),
							   width, height);
    ASRL_CHECK_CUDA_ERROR_DBG("transpose 1");    

    // compute row scan on transposed image (columns)
    cudppMultiScan(colPlan, trans_data, trans_data, height, width);
    ASRL_CHECK_CUDA_ERROR_DBG("column scan");


    // transpose (again, to return to original orientation)
    grid.x = (int) ((block.x + height - 1) / block.x);
    grid.y = (int) ((block.y + width - 1) / block.y);

    transpose_kernel<float> <<< grid, block, 0 >>> (norm_data, norm_pitch/sizeof(float), trans_data, trans_pitch/sizeof(float),
						    height, width);
    ASRL_CHECK_CUDA_ERROR_DBG("transpose 2");

    // compute row scan on re-transposed image (rows)
    cudppMultiScan(rowPlan, norm_data, norm_data, width, height);
    ASRL_CHECK_CUDA_ERROR_DBG("row scan");

    // transfer data to cudaArray
    cudaError_t err = cudaMemcpy2DToArray(output_int_img, 0, 0, norm_data, norm_pitch, width*sizeof(float), height,
					cudaMemcpyDeviceToDevice);
    ASRL_ASSERT_EQ(err,cudaSuccess, "Unable to copy the integral image to the texture buffer: (" << err << "): " << cudaGetErrorString(err));		
  }


} // namespace asrl


